@@ -7,15 +7,13 @@ use Readonly;
 use English '-no_match_vars';
 
 require Exporter;
-use AutoLoader qw(AUTOLOAD);
 use base 'Exporter';
-use Carp;
-use Socket;
-use IO::Handle;
+
+use IO::Socket::INET;
 use Net::Gnats::PR;
 use Net::Gnats::Response;
 
-our $VERSION = '0.11';
+our $VERSION = '0.13';
 our @CARP_NOT;
 $OUTPUT_AUTOFLUSH = 1;
 
@@ -204,14 +202,43 @@ sub new {
     $self->{errorMessage} = undef;
     $self->{accessMode} = undef;
     $self->{gnatsdVersion} = undef;
+    $self->{skip_version_check} = 0;
     $self->{user} = undef;
     $self->{db}   = undef;
 
     return $self;
 }
 
+sub gnatsd_version {
+  my ($self, $value) = @_;
+  if (defined $value) {
+    $value =~ s/.*(\d+.\d+.\d+).*/$1/;
+    $self->{gnatsdVersion} = $1;
+  }
+  return $self->{gnatsdVersion};
+}
+
+# "legally" use v4 daemon only
+sub check_gnatsd_version {
+  my ($self) = @_;
+  my $rmajor = 4;
+  my $min_minor = 1;
+  return 1 if $self->skip_version_check;
+  my ($majorv, $minorv, $patchv) = split /\./, $self->gnatsd_version;
+
+  return 0 if $majorv != $rmajor;
+  return 0 if $minorv < $min_minor;
+  return 1;
+}
+
+sub skip_version_check {
+  my ($self, $value) = @_;
+  $self->{skip_version_check} = $value if defined $value;
+  return $self->{skip_version_check};
+}
+
 sub debug_gnatsd {
-  $debug_gnatsd = 1;
+  $debug_gnatsd = 0;
   return;
 }
 
@@ -227,72 +254,60 @@ sub gnatsd_connect {
     my ( $self ) = @_;
     my ( $sock, $iaddr, $paddr, $proto );
 
-    #TODO disconnect if already connected
+    $self->disconnect if defined $self->_gsock;
 
-    if ( ! ( $iaddr = inet_aton($self->{hostAddr} ) ) ) {
-        carp("Unknown GNATS host '$self->{hostAddr}'");
-        return 0;
+    my $socket = IO::Socket::INET->new( PeerAddr => $self->{ hostAddr },
+                                        PeerPort => $self->{ hostPort },
+                                        Proto    => 'tcp');
+
+    if ( not defined $socket ) {
+      debug('$socket is not defined.');
+      return;
     }
 
-    $paddr = sockaddr_in( $self->{ hostPort }, $iaddr);
-    $proto = getprotobyname 'tcp' ;
-    if ( not socket $sock, PF_INET, SOCK_STREAM, $proto ) {
-      #TODO: RECOVER BETTER HERE
-      carp "gnatsweb: client_init error $self->{hostAddr} $self->{hostPort}: $OS_ERROR";
-      return 0;
-    }
+    $self->_gsock( $socket );
 
-    $self->_gsock( $sock );
-
-    if ( not connect $self->_gsock, $paddr ) {
-      #TODO: RECOVER BETTER HERE
-      carp "gnatsweb: client_init error $self->{hostAddr} $self->{hostPort}: $OS_ERROR ;";
-      return 0;
-    }
-
-    $self->_gsock->autoflush(1);
     my $response = $self->_get_gnatsd_response();
-    $self->{lastCode} = $response->code;
+
+    $self->{lastCode} = defined $response->code ? $response->code : undef;
     $self->{lastResponse} = $response->as_string;
     debug('INIT: [' . $response->as_string . ']');
 
     # Make sure we got a 200 code.
     if ($response->code != $CODE_GREETING) {
-      carp "? Error: Unknown gnatsd connection response: $response";
+      logerror("Unknown gnatsd connection response: " . $response->as_string);
       return 0;
     }
 
     # Grab the gnatsd version
-    if ( $response->raw =~ /\d.\d.\d/sxm ) {
-      $self->{gnatsdVersion} = $response->raw;
-      $self->{gnatsdVersion} =~ s/.*(\d.\d.\d).*/$1/sxm;
-    }
-    else {
+    $self->gnatsd_version( pop @{ $response->raw } );
+    if (not $self->check_gnatsd_version) {
       # We only know how to talk to gnats4
       warn "? Error: GNATS Daemon version $self->{gnatsdVersion} at $self->{hostAddr} $self->{hostPort} is not supported by Net::Gnats\n";
       return 0;
     }
+
     $self->{newPRs} = 0;
     return 1;
 }
 
 sub disconnect {
     my ( $self ) = @_;
-    return $self->_do_gnats_cmd('QUIT');
+    $self->_do_gnats_cmd('QUIT');
+    $self->{sock}->close;
+    $self->{sock} = undef;
+    return 1;
 }
 
 sub get_dbnames {
     my ( $self ) = @_;
 
-    my ($code, $response) = $self->_do_gnats_cmd('DBLS');
-    debug('DBLS CODE: [' . $code . ']');
+    my $r = $self->_do_gnats_cmd('DBLS');
+    debug('DBLS CODE: [' . $r->code . ']');
 
-    if ($self->_is_code_ok($code)) {
-        return $self->_extract_list_content($response);
-    }
+    return $r->raw if $r->code == $CODE_TEXT_READY;
 
-    debug('DBLS DID NOT PASS IS_CODE_OK');
-    $self->_mark_error($code, $response);
+    $self->_mark_error($r);
     return;
 }
 
@@ -310,8 +325,8 @@ sub list_categories {
 
 sub list_submitters {
     return shift->_list('SUBMITTERS',
-                        ['name', 'desc', 'contract', 'something1',
-                         'responsible']);
+                        ['name', 'desc', 'contract', 'response',
+                         'contact', 'othernotify']);
 }
 
 sub list_responsible {
@@ -326,49 +341,56 @@ sub list_states {
 
 sub list_fieldnames {
   my ( $self ) = @_;
-  my ($code, $response) = $self->_do_gnats_cmd('LIST FIELDNAMES');
-  return if not $self->_is_code_ok($code);
-  return $response;
+  my $r = $self->_do_gnats_cmd('LIST FIELDNAMES');
+  return $r->raw if $r->code == $CODE_TEXT_READY;
+  return;
 }
 
 sub list_inputfields_initial {
   my ( $self ) = @_;
+
   if ($#{$self->{fieldData}->{initial}} < 0) {
-    my ($code, $response) = $self->_do_gnats_cmd('LIST INITIALINPUTFIELDS');
-    if ($self->_is_code_ok($code)) {
-      push @{$self->{fieldData}->{initial}}, $self->_extract_list_content($response);
-    } else {
-      $self->_mark_error($code, $response);
+    my $r = $self->_do_gnats_cmd('LIST INITIALINPUTFIELDS');
+
+    if ($r->code != $CODE_TEXT_READY) {
+      $self->_mark_error($r);
       return;
     }
+
+    push @{$self->{fieldData}->{initial}}, @{$r->raw};
+  }
+  return $self->{fieldData}->{initial};
+}
+
+sub list_inputfields_initial_required {
+  my ( $self ) = @_;
+
+  if ($#{$self->{fieldData}->{initial}} < 0) {
+    my $r = $self->_do_gnats_cmd('LIST INITIALREQUIREDFIELDS');
+
+    if ($r->code != $CODE_TEXT_READY) {
+      $self->_mark_error($r);
+      return;
+    }
+
+    push @{$self->{fieldData}->{initial}}, @{$r->raw};
   }
   return wantarray ? @{$self->{fieldData}->{initial}} : $self->{fieldData}->{initial};
 }
 
+
+
 sub get_field_type {
   my ( $self, $field ) = @_;
 
   if (not defined $field) { return; }
 
+  my $r = $self->_do_gnats_cmd("FTYP $field");
 
+  if ( $r->code == $CODE_INVALID_FIELD_NAME ) { return; }
 
-#  if ( defined $self->{fieldData}->{fields}->{$field}->{type} ) {
-#    return $self->{fieldData}->{fields}->{$field}->{type};
-#  }
-
-  my ($code, $response) = $self->_do_gnats_cmd("FTYP $field");
-  if ( $code == $CODE_INVALID_FIELD_NAME ) { return; }
-
-#  if ( ! $self->_is_code_ok($code) ) {
-#    $self->_mark_error($code, $response);
-#    return;
-#  }
-
-#  $response =~ s/^\d+\s+//sxm;
-#  chomp $response;
-#  $self->{fieldData}->{fields}->{$field}->{type} = $response;
-
-  return $response;
+  # only one value should be returned
+  return shift @{ $r->raw };
 }
 
 sub is_validfield {
@@ -383,78 +405,71 @@ sub get_field_typeinfo {
   my ( $self, $field, $property ) = @_;
   if ( not defined $field ) { return; }
   my $type_response = $self->get_field_type($field);
+
   debug('FTYP (response): [' . $type_response . ']');
+
   if ( $type_response ne 'MultiEnum' ) { return; }
   if ( not defined $property ) { $property = 'separators'; }
 
 
-  my ($code, $response) = $self->_do_gnats_cmd("FTYPINFO $field $property");
-  if ( $code == $CODE_INVALID_FTYPE_PROPERTY ) { return; }
-
-  # if (not exists($self->{fieldData}->{fields}->{$field}->{typeInfo})) {
-  #   if ($self->_is_code_ok($code)) {
-  #     $self->{fieldData}->{fields}->{$field}->{typeInfo} = $response;
-  #   } else {
-  #     $self->_mark_error($code, $response);
-  #     return;
-  #   }
-  # }
-  # return $self->{fieldData}->{fields}->{$field}->{typeInfo};
-  return $response;
+  my $r = $self->_do_gnats_cmd("FTYPINFO $field $property");
+  if ( $r->code == $CODE_INVALID_FTYPE_PROPERTY ) { return; }
+  return $r->raw;
 }
 
 sub get_field_desc {
-    my ( $self, $field ) = @_;
+  my ( $self, $field ) = @_;
 
-    my ($code, $response) = $self->_do_gnats_cmd("FDSC $field");
-    if ($self->_is_code_ok($code)) {
-        return $response;
-    } else {
-        $self->_mark_error($code, $response);
-        return;
-    }
+  my $r = $self->_do_gnats_cmd("FDSC $field");
+  return shift @{ $r->raw } if $r->code == $CODE_INFORMATION;
+
+  $self->_mark_error($r);
+  return;
 }
 
 sub get_field_flags {
   my ( $self, $field, $flag ) = @_;
+
   if ( not defined $field ) { return; }
 
-  my ($code, $response) = $self->_do_gnats_cmd("FIELDFLAGS $field");
+  my $r = $self->_do_gnats_cmd("FIELDFLAGS $field");
 
-  if (defined $flag and $response =~ /$flag/sxm) { return 1; }
+  if (defined $flag and $r->raw =~ /$flag/sxm) { return 1; }
 
-  return $response;
+  return $r->raw;
 }
 
 sub get_field_validators {
     my ( $self, $field ) = @_;
 
-    my ($code, $response) = $self->_do_gnats_cmd("FVLD $field");
+    return if not defined $field;
 
-    if ($self->_is_code_ok($code)) {
-        my @validators = $self->_extract_list_content($response);
-        return @validators;
-    } else {
-        $self->_mark_error($code, $response);
-        return;
+    my $r = $self->_do_gnats_cmd("FVLD $field");
+
+    return $r->raw if $r->code == $CODE_TEXT_READY;
+
+    if ( $r->code == $CODE_INVALID_FIELD_NAME ) {
+      $self->_mark_error($r);
+      return;
     }
+
+    return;
 }
 
 
 sub validate_field {
   my ( $self, $field, $input ) = @_;
 
-  my ($code, $response) = $self->_do_gnats_cmd("VFLD $field");
+  return if not defined $field or not defined $input;
 
-  if ( $code != $CODE_SEND_TEXT ) { # MUST respond with 212 to continue.
-    carp('ERROR: [' . $code . '] when calling VFLD on [' . $field . ']');
-    return;
-  }
+  my $r = $self->_do_gnats_cmd("VFLD $field");
 
-  ($code, $response) = $self->_do_gnats_cmd($input . $NL . q{.});
+  return if $r->code != $CODE_SEND_TEXT;
 
-  if ( $code != $CODE_OK ) {
-    carp('ERROR: [' . $code . '] when supplying VFLD text on [' . $field . ']');
+  $r = $self->_do_gnats_cmd($input . $NL . q{.});
+
+  if ( $r->code != $CODE_OK ) {
+    logerror('ERROR: [' . $r->code . '] when supplying VFLD text on [' . $field . ']');
     return;
   }
 
@@ -464,21 +479,21 @@ sub validate_field {
 
 sub get_field_default {
   my ( $self, $field ) = @_;
-  my ($code, $response) = $self->_do_gnats_cmd("INPUTDEFAULT $field");
-  if (not $self->_is_code_ok($code)) { return }
-  return @{ $response }[0];
+  my $r = $self->_do_gnats_cmd("INPUTDEFAULT $field");
+  return if $r->code != $CODE_OK;
+  return @{ $r->raw }[0];
 }
 
 
 sub reset_server {
   my ( $self ) = @_;
 
-  my ($code, $response) = $self->_do_gnats_cmd('RSET');
+  my $r = $self->_do_gnats_cmd('RSET');
 
   # CODE_CMD_ERROR (440) can never happen since we constrain no args in code.
 
-  if ( $code >= $CODE_ERROR ) {
-    carp( 'ERROR [' . $code . ']: ' . $response );
+  if ( $r->code >= $CODE_ERROR ) {
+    logerror( 'ERROR [' . $r->code . ']: ' . $r->raw );
     return;
   }
 
@@ -489,42 +504,41 @@ sub reset_server {
 sub lock_main_database {
   my ( $self ) = @_;
 
-  my ($code, $response) = $self->_do_gnats_cmd('LKDB');
+  my $r = $self->_do_gnats_cmd('LKDB');
 
-  # CODE_CMD_ERROR (440) can never happen since we constrain no args in code.
+  logerror('ERROR: CODE_GNATS_LOCKED: Gnats database already locked.')
+    and return
+    if $r->code == $CODE_GNATS_LOCKED;
 
-  if ( $code == $CODE_GNATS_LOCKED ) {
-    carp('ERROR: CODE_GNATS_LOCKED: Gnats database already locked.');
-    return;
-  }
+  logerror('ERROR: CODE_CMD_ERROR')
+    and return
+    if $r->code == $CODE_CMD_ERROR;
 
-  # Test this with a user who does not have privelege to lock database.
-  if ( $code >= $CODE_ERROR ) {
-    carp( 'ERROR [' . $code . ']: ' . $response );
-    return;
-  }
+  logerror( 'ERROR [' . $r->code . ']: ' . $r->raw )
+    and return
+    if $r->code >= $CODE_ERROR;
+
+  return if $r->code == -1;
 
   return 1;
 }
 
 sub unlock_main_database {
-    my ( $self ) = @_;
-    my $code; my $response;
-    ($code, $response) = $self->_do_gnats_cmd('UNDB');
+  my ( $self ) = @_;
 
+  my $r = $self->_do_gnats_cmd('UNDB');
 
-  # CODE_CMD_ERROR (440) can never happen since we constrain no args in code.
+  # CODE_CMD_ERROR (440) can never happen since we constrain no args
+  # in code.
+  logerror('ERROR: CODE_GNATS_NOT_LOCKED: Gnats database not locked.')
+    and return
+    if $r->code == $CODE_GNATS_NOT_LOCKED;
 
-  if ( $code == $CODE_GNATS_NOT_LOCKED ) {
-    carp('ERROR: CODE_GNATS_NOT_LOCKED: Gnats database not locked.');
-    return;
-  }
-
-  # Test this with a user who does not have privelege to lock database.
-  if ( $code >= $CODE_ERROR ) {
-    carp( 'ERROR [' . $code . ']: ' . $response );
-    return;
-  }
+  # Test this with a user who does not have privelege to lock
+  # database.
+  logerror( 'ERROR [' . $r->code . ']: ' . $r->raw )
+    and return
+    if $r->code >= $CODE_ERROR;
 
   return 1;
 }
@@ -532,100 +546,88 @@ sub unlock_main_database {
 sub lock_pr {
   my ( $self, $pr_number, $user ) = @_;
 
-  my ($code, $response) = $self->_do_gnats_cmd("LOCK $pr_number $user");
+  return if not defined $pr_number or not defined $user;
 
-  if ( $code == $CODE_CMD_ERROR ) {
-    carp( 'ERROR: CODE_CMD_ERROR: ' . $response );
-    return;
-  }
+  my $r = $self->_do_gnats_cmd("LOCK $pr_number $user");
 
-  if ( $code == $CODE_NONEXISTENT_PR ) {
-    carp( 'ERROR: CODE_NONEXISTENT_PR: ' . $response );
-    return;
-  }
+  logerror( 'ERROR: CODE_CMD_ERROR: ' . $r->raw )
+    and return
+    if $r->code == $CODE_CMD_ERROR;
 
-  if ( $code == $CODE_LOCKED_PR ) {
-    carp( 'ERROR: CODE_LOCKED_PR: ' . $response );
-    return;
-  }
+  logerror( 'ERROR: CODE_NONEXISTENT_PR: ' . $r->raw )
+    and return
+    if $r->code == $CODE_NONEXISTENT_PR;
+
+  logerror( 'ERROR: CODE_LOCKED_PR: ' . $r->raw )
+    and return
+    if $r->code == $CODE_LOCKED_PR;
 
   # Test this with a user who does not have privelege to lock the PR.
-  if ( $code >= $CODE_ERROR ) {
-    carp( 'ERROR [' . $code . ']: ' . $response );
-    return;
-  }
+  logerror( 'ERROR [' . $r->code . ']: ' . $r->raw )
+    and return
+    if $r->code >= $CODE_ERROR;
 
   # CODE_PR_READY (300)
   my $pr = Net::Gnats::PR->new( $self );
-  $pr->parse( $response );
+  $pr->parse( $r->raw );
   return $pr;
 }
 
 sub unlock_pr {
   my ( $self, $pr ) = @_;
 
-  if ( not defined $pr ) {
-    carp( 'ERROR: unlock_pr requires PR number' );
-    return;
-  }
+  logerror( 'ERROR: unlock_pr requires PR number' )
+    and return
+    if not defined $pr;
 
-  my ($code, $response) = $self->_do_gnats_cmd( 'UNLK ' . $pr );
+  my $r = $self->_do_gnats_cmd( 'UNLK ' . $pr );
 
-  if ( $code == $CODE_CMD_ERROR ) {
-    carp( 'ERROR: CODE_CMD_ERROR: ' . $response );
-    return;
-  }
+  return 1 if $r->code == $CODE_OK;
 
-  if ( $code == $CODE_NONEXISTENT_PR ) {
-    carp( 'ERROR: CODE_NONEXISTENT_PR: ' . $response );
-    return;
-  }
+  logerror( 'ERROR: CODE_CMD_ERROR: ' . $r->raw)
+    and return
+    if $r->code == $CODE_CMD_ERROR;
 
-  if ( $code == $CODE_PR_NOT_LOCKED ) {
-    carp( 'ERROR: CODE_PR_NOT_LOCKED: ' . $response );
-    return;
-  }
+  logerror( 'ERROR: CODE_NONEXISTENT_PR: ' . $r->raw )
+    and return
+    if $r->code == $CODE_NONEXISTENT_PR;
+
+  logerror( 'ERROR: CODE_PR_NOT_LOCKED: ' . $r->raw )
+    and return
+    if $r->code == $CODE_PR_NOT_LOCKED;
 
   # Test this with a user who does not have privelege to lock the PR.
-  if ( $code >= $CODE_ERROR ) {
-    carp( 'ERROR [' . $code . ']: ' . $response );
-    return;
-  }
+  logerror( 'ERROR [' . $r->code . ']: ' . $r->raw )
+    and return
+    if $r->code >= $CODE_ERROR;
 
-  # CODE_OK
-  return 1;
+  return;
 }
 
 sub delete_pr {
   my ( $self, $pr ) = @_;
 
-  my ($code, $response) = $self->_do_gnats_cmd('DELETE ' . $pr->getField('Number'));
+  my $r = $self->_do_gnats_cmd('DELETE ' . $pr->getField('Number'));
 
-  if ( $code == $CODE_OK ) {
-    return 1;
-  }
+  return 1 if $r->code == $CODE_OK;
 
-  if ( $code == $CODE_NO_ACCESS ) {
-    carp('You do not have access to delete this PR.');
-    return;
-  }
+  logerror('You do not have access to delete this PR.')
+    and return
+    if $r->code == $CODE_NO_ACCESS;
 
-  if ( $code == $CODE_LOCKED_PR ) {
-    carp('You cannot delete a locked PR.');
-    return;
-  }
+  logerror('You cannot delete a locked PR.')
+    and return
+    if $r->code == $CODE_LOCKED_PR;
 
-  if ( $code == $CODE_GNATS_LOCKED ) {
-    carp('Cannot delete, Gnats DB is currently locked.');
-    return;
-  }
+  logerror('Cannot delete, Gnats DB is currently locked.')
+    and return
+    if $r->code == $CODE_GNATS_LOCKED;
 
-  if ( $code == $CODE_NONEXISTENT_PR ) {
-    carp('PR nonexistent.');
-    return;
-  }
+  logerror('PR nonexistent.')
+    and return
+    if $r->code == $CODE_NONEXISTENT_PR;
 
-  carp('Unexpected error [' . $code . '] occurred. PR not deleted.');
+  logerror('Unexpected error [' . $r->code . '] occurred. PR not deleted.');
   return;
 }
 
@@ -635,50 +637,52 @@ sub check_newpr {
   return;
 }
 
+sub chek {
+  my ( $self, $initial ) = @_;
+
+  $initial = defined $initial ? 'initial' : '';
+
+  my $r = $self->_do_gnats_cmd("CHEK $initial");
+
+  # TODO: Add logging
+  return 1 if $r->code == $CODE_SEND_PR;
+
+  # TODO: Add logging
+  return undef if $r->code == $CODE_CMD_ERROR;
+
+  logerror('Unexpected error [' . $r->code . '] occurred. PR not deleted.');
+  return;
+}
+
+
 sub check_pr {
   my ( $self, $pr, $arg ) = @_;
 
   my $argument  = defined $arg ? $arg : q{};
 
-  my ($code, $response) = $self->_do_gnats_cmd("CHEK $argument");
+  my $r = $self->_do_gnats_cmd("CHEK $argument");
 
-  if ($self->_is_code_ok($code)) {
-    ($code, $response) = $self->_do_gnats_cmd( $pr . $NL . $DOT );
-    if ($self->_is_code_ok($code)) {
-      return 1;
-    }
-    # TODO: Should this be handled in _doGnatsCmd?
-    # If gnatsd returns a: 401 Couldn't read PR header
-    # or if it sends multiple: 411 There is a bad value ... for field ...
-    # then it also sends a: 403 Errors found checking PR text
-    # So read back that response, but ignore it cause the user
-    # would want to see the 401 error
-    my @response = split $NL, $response;
+  return if $r->code != $CODE_SEND_PR;
 
-    # TODO This condition is moderately retarded
-    if (   $code == $CODE_INVALID_ENUM or
-         ( $code == $CODE_INVALID_ENUM and $#response > 0)) {
-      my ($rcode, $rresponse) = $self->_process;
-    }
-    $self->_mark_error($code, $response);
-    return 0;
-  }
+  $r = $self->_do_gnats_cmd( $pr . $NL . $DOT );
 
-  $self->_mark_error($code, $response);
-  return 0;
+  return 1 if $r->code == $CODE_OK;
+
+  # TODO: If at this point, there can be "INNER ERRORS" which need to
+  # be captured and reported on via Net::Gnats::Response.
+  return;
 }
 
 
 sub set_workingemail {
-    my ( $self, $email ) = @_;
+  my ( $self, $email ) = @_;
 
-    my ($code, $response) = $self->_do_gnats_cmd("EDITADDR $email");
-    if ($self->_is_code_ok($code)) {
-        return 1;
-    } else {
-        $self->_mark_error($code, $response);
-        return 0;
-    }
+  my $r = $self->_do_gnats_cmd("EDITADDR $email");
+
+  return 1 if $r->code == $CODE_OK;
+
+  $self->_mark_error($r)
+    and return;
 }
 
 #
@@ -688,42 +692,51 @@ sub set_workingemail {
 
 sub truncate_field_content {
   my ( $self, $pr, $field, $input, $reason ) = @_;
-  confess '? Error: pr not passed to replaceField' if not defined $pr;
-  confess '? Error: field passed to replaceField' if not defined $field;
-  confess '? Error: no input passed to replaceField' if not defined $input;
+  logerror('? Error: pr not passed to replaceField')
+    if not defined $pr;
+
+  logerror('? Error: field passed to replaceField')
+    if not defined $field;
+
+  logerror('? Error: no input passed to replaceField')
+    if not defined $input;
 
   # See if this field requires a change reason.
   # TODO: We could just enter the $input, and see if gnatsd says
   #       a reason is required, but I could not figure out how to
   #       abort at that point if no reason was given...
   my $need_reason = $self->getFieldFlags($field, 'requireChangeReason');
+
   if ($need_reason and ( not defined $reason or $reason eq q{} )) {
-    $self->_mark_error($CODE_INVALID_PR_CONTENTS,
-                       'No change Reason Specified');
-    return(0);
+    logerror('No change Reason Specified');
+    return;
   }
 
-  my ($code, $response) = $self->_do_gnats_cmd("REPL $pr $field");
+  my $r = $self->_do_gnats_cmd("REPL $pr $field");
+
+  if ( $r->code == $CODE_SEND_TEXT ) {
+    $r = $self->_do_gnats_cmd($input . $NL . $DOT);
 
-  if ($self->_is_code_ok($code)) {
-    ($code, $response) = $self->_do_gnats_cmd($input . $NL . $DOT);
     if ($need_reason) {
       #warn "reason=\"$reason\"";
       # TODO: This can choke here if we encounter a PR with a bad field like:
       # _getGnatsdResponse: READ >>411 There is a bad value `unknown' for the field `Category'.
-      ($code, $response) = $self->_do_gnats_cmd($reason . $NL . $DOT)
+      $r = $self->_do_gnats_cmd($reason . $NL . $DOT)
     }
-    if ($self->_is_code_ok($code)) {
+
+    $self->restart($r->code)
+      and return $self->replaceField($pr, $field, $input, $reason)
+      if $r->code == $CODE_FILE_ERROR;
+
+    if ($self->_is_code_ok($r->code)) {
       return 1;
     }
-    $self->_mark_error($code, $response);
-  } else {
-    $self->_mark_error($code, $response);
+    $self->_mark_error($r);
+
   }
-  if ($code == $CODE_FILE_ERROR and $self->restart($code)) {
-    return $self->replaceField($pr, $field, $input, $reason);
-  }
-  return 0;
+
+  $self->_mark_error($r );
+  return;
 }
 
 my $restart_time;
@@ -734,12 +747,12 @@ sub restart {
   my $ctime = time;
   if ( defined $restart_time ) {
     if ( ($ctime - $restart_time) < $RESTART_CHECK_THRESHOLD ) {
-      carp '! ERROR: Restart attempted twice in a row, 640 error must be real!';
+      logerror('! ERROR: Restart attempted twice in a row, 640 error must be real!');
       return 0;
     }
   }
 
-  carp ( $NL
+  logerror ( $NL
       .  $NL . '! ERROR: Recieved GNATSD code ' . $code . ', will now disconnect and'
       .  $NL . 'reconnecting to gnatsd, then re-issue the command.  This may cause any'
       .  $NL . 'following commands to behave differently if you depended on'
@@ -756,29 +769,32 @@ sub restart {
 }
 
 sub append_field_content {
-    my ( $self, $pr, $field, $input ) = @_;
+  my ( $self, $pr, $field, $input ) = @_;
 
-    confess '? Error: pr not passed to appendField' if not defined $pr;
-    confess '? Error: field passed to appendField' if not defined $field;
-    confess '? Error: no input passed to appendField' if not defined $input;
+  logerror('? Error: pr not passed to appendField')
+    if not defined $pr;
+  logerror('? Error: field passed to appendField')
+      if not defined $field;
+  logerror('? Error: no input passed to appendField')
+    if not defined $input;
 
-    my ($code, $response) = $self->_do_gnats_cmd("APPN $pr $field");
+  my $r = $self->_do_gnats_cmd("APPN $pr $field");
 
-    if ($self->_is_code_ok($code)) {
-        ($code, $response) = $self->_do_gnats_cmd( $input . $NL . $DOT );
-        if ($self->_is_code_ok($code)) {
-            return 1;
-        } else {
-            $self->_mark_error($code, $response);
-        }
+  if ($self->_is_code_ok($r->code)) {
+    $r= $self->_do_gnats_cmd( $input . $NL . $DOT );
+    if ($self->_is_code_ok($r->code)) {
+      return 1;
     } else {
-        $self->_mark_error($code, $response);
+      $self->_mark_error( $r );
     }
-    if ($code == $CODE_FILE_ERROR and $self->restart($code)) {
-      # TODO: This can potentially be an infinte loop...
-      return $self->appendToField($pr,$field,$input);
-    }
-    return 0;
+  } else {
+    $self->_mark_error($r);
+  }
+  if ($r->code == $CODE_FILE_ERROR and $self->restart($r->code)) {
+    # TODO: This can potentially be an infinte loop...
+    return $self->appendToField($pr, $field, $input);
+  }
+  return 0;
 }
 
 sub submit_pr {
@@ -790,25 +806,25 @@ sub submit_pr {
 
   my $pr_string = $pr->unparse();
 
-  my ($code, $response) = $self->_do_gnats_cmd('SUBM');
+  my $r = $self->_do_gnats_cmd('SUBM');
 
-  if ($code == $CODE_GNATS_LOCKED) {
-    carp( 'Gnats database locked, cannot submit PR.' );
+  if ($r->code == $CODE_GNATS_LOCKED) {
+    logerror( 'Gnats database locked, cannot submit PR.' );
     return;
   }
 
-  ($code, $response) = $self->_do_gnats_cmd($pr_string . $NL . q{.});
+  $r = $self->_do_gnats_cmd($pr_string . $NL . q{.});
 
   # Returns PR Number. Return this to the caller.
-  if ( $code == $CODE_INFORMATION or
-       $code == $CODE_INFORMATION_FILLER ) {
+  if ( $r->code == $CODE_INFORMATION or
+       $r->code == $CODE_INFORMATION_FILLER ) {
     $self->{newPRs}++;
-    return $response;
+    return $r->raw;
   }
 
   # Something unexpected happened.  The client can attempt to resend.
   # Later, give the client the whole response object.
-  carp('ERROR: Unexpected response code [' . $code . ']: ' . @{ $response }[0]);
+  logerror('ERROR: Unexpected response code [' . $r->code . ']: ' . @{ $r->raw }[0]);
   return;
 }
 
@@ -844,38 +860,35 @@ sub update_pr {
   $slast_modified ||= q{}; # Default to empty
 
   if ($last_modified ne $slast_modified) {
-    carp('Someone modified the PR. Refresh the PR and try again.');
+    logerror('Someone modified the PR. Refresh the PR and try again.');
     return;
   }
 
-  ($code, $response) = $self->_do_gnats_cmd('EDITADDR ' . $self->{user});
-  if ( $code == $CODE_CMD_ERROR ) {
-    carp('ERROR: EDITADDR: ' . $response);
-    return;
-  }
+  my $r = $self->_do_gnats_cmd('EDITADDR ' . $self->{user});
 
-  ($code, $response) = $self->_do_gnats_cmd('EDIT ' . $pr->getField('Number'));
+  logerror('ERROR: EDITADDR: ' . $r->raw)
+    and return
+    if $r->code == $CODE_CMD_ERROR;
 
-  if ( $code == $CODE_GNATS_LOCKED ) {
-    carp('ERROR: EDIT: CODE_GNATS_LOCKED: ' . $response);
-    return;
-  }
-  if ( $code == $CODE_PR_NOT_LOCKED ) {
-    carp('ERROR: EDIT: CODE_PR_NOT_LOCKED: ' . $response);
-    return;
-  }
-  if ( $code == $CODE_NONEXISTENT_PR ) {
-    carp('ERROR: EDIT: CODE_NONEXISTENT_PR: ' . $response);
-    return;
-  }
-  # Assume CODE_SEND_PR response at this point.
+  $r = $self->_do_gnats_cmd('EDIT ' . $pr->getField('Number'));
 
-  ($code, $response) = $self->_do_gnats_cmd( $pr_string . q{.} );
+  logerror('ERROR: EDIT: CODE_GNATS_LOCKED: ' . $r->raw)
+    and return
+    if $r->code == $CODE_GNATS_LOCKED;
 
-  if ( $code != $CODE_OK ) {
-    carp('ERROR: EDIT: FILING FAILED: ' . $response);
-    return;
-  }
+  logerror('ERROR: EDIT: CODE_PR_NOT_LOCKED: ' . $r->raw)
+    and return
+    if $r->code == $CODE_PR_NOT_LOCKED;
+
+  logerror('ERROR: EDIT: CODE_NONEXISTENT_PR: ' . $r->raw)
+    and return
+    if $r->code == $CODE_NONEXISTENT_PR;
+
+  $r = $self->_do_gnats_cmd( $pr_string . q{.} );
+
+  logerror('ERROR: EDIT: FILING FAILED: ' . $r->raw)
+    and return
+    if $r->code != $CODE_OK;
 
   $self->unlock_pr($pr->getField('Number'));
 
@@ -885,17 +898,14 @@ sub update_pr {
 
 sub new_pr {
   my ( $self ) = @_;
-  return Net::Gnats::PR->new($self);
-}
 
-# Fillout all defaults for a PR.
-# TODO: should this always be called by newPR?
-sub fillout_pr {
-  my ( $self, $pr ) = @_;
+  my $pr = Net::Gnats::PR->new($self);
+
   foreach my $field ($self->listInitialInputFields) {
-    $pr->setField($field,$self->getFieldDefault($field));
+    $pr->setField($field,
+                  $self->getFieldDefault( $field ) );
   }
-  return;
+  return $pr;
 }
 
 sub get_pr_by_number {
@@ -905,95 +915,94 @@ sub get_pr_by_number {
     return;
   }
 
-  my ($code, $response) = $self->_do_gnats_cmd('QFMT full');
-  if (not $self->_is_code_ok($code)) {
-    $self->_mark_error($code, $response);
-    return;
-  }
+  my $r = $self->_do_gnats_cmd('QFMT full');
 
-  ($code, $response) = $self->_do_gnats_cmd("QUER $num");
+  $self->_mark_error($r)
+    and return
+    if not $self->_is_code_ok($r->code);
 
-  debug('CODE: ' . $code );
-  debug('RESPONSE: ' . @{ $response }[0] );
-  if ( $code == $CODE_NO_PRS_MATCHED and
-       @{$response}[0] eq 'No PRs match.' ) {
-    return;
-  }
+  $r = $self->_do_gnats_cmd("QUER $num");
 
-  if (not $self->_is_code_ok($code)) {
-    $self->_mark_error($code, $response);
-    return;
-  }
+  debug('CODE: ' . $r->code );
+  debug('RESPONSE: ' . @{ $r->raw }[0] );
+
+  return if $r->code == $CODE_NO_PRS_MATCHED;
+
+
+  $self->_mark_error($r)
+    and return
+    if not $self->_is_code_ok($r->code);
 
   my $pr = $self->new_pr();
-  $pr->parse( @{ $response } ) ;
+  $pr->parse( @{ $r->raw } ) ;
 
   return $pr;
 }
 
 
 sub expr {
-    my ( $self ) = @_;
-    my @exprs = @_;
+  my $self = shift;
+  my @exprs = @_;
+  return if scalar @exprs == 0;
 
-    my ($code, $response);
-    foreach my $expr (@exprs) {
-        ($code, $response) = $self->_do_gnats_cmd("EXPR $expr");
-    }
-    return $code; #XXX TODO and codes together or abort or something
+  foreach my $expr (@exprs) {
+    my $r = $self->_do_gnats_cmd("EXPR $expr");
+    return if $r->code == $CODE_INVALID_EXPR;
+  }
+
+  return 1;
+}
+
+# Because we don't know what's in the dbconfig file, we will only
+# support FULL, STANDARD, and SUMMARY since those must be defined.
+# Otherwise, we assume it is a custom format.
+sub qfmt {
+  my ($self, $format) = @_;
+
+  # If format is not defined, then defaults to STANDARD
+  # This is per the GNATS specification.
+  $format = 'standard' if not defined $format;
+
+  my $r = $self->_do_gnats_cmd("QFMT $format");
+
+  return 1 if $r->code == $CODE_OK;
+  return   if $r->code == $CODE_CMD_ERROR;
+  return   if $r->code == $CODE_INVALID_QUERY_FORMAT;
+  return;
 }
 
 sub query {
-  my ( $self ) = @_;
+  my $self = shift;
   my @exprs = @_;
 
-  if ( not defined $self->reset_server ) {
-    return;
-  }
+  return if not defined $self->reset_server;
+  return if not defined $self->qfmt('full');
+  return if not defined $self->expr(@exprs);
 
-  my ($code, $response) = $self->_do_gnats_cmd('QFMT Number'); #XXX TODO
-  if (not $self->_is_code_ok($code)) {
-    $self->_mark_error($code, $response);
-    return;
-  }
-
-  foreach my $expr (@exprs) {
-    ($code, $response) = $self->_do_gnats_cmd("EXPR $expr");
-    if (not $self->_is_code_ok($code)) {
-      $self->_mark_error($code, $response);
-      return;
-    }
-  }
-
-  my @nums;
-  ($code, $response) = $self->_do_gnats_cmd('QUER');
-  if ($self->_is_code_ok($code)) {
-    @nums = $self->_extract_list_content($response);
-  } else {
-    $self->_mark_error($code, $response);
-    return;
-  }
-  return @nums;
+  my $r = $self->_do_gnats_cmd('QUER');
+  return $r->raw if $r->code == $CODE_PR_READY;
+  return []      if $r->code == $CODE_NO_PRS_MATCHED;
+  return         if $r->code == $CODE_INVALID_QUERY_FORMAT;
+  return;
 }
 
 sub _list {
   my ( $self, $listtype, $keynames ) = @_;
 
-  my ($code, $response) = $self->_do_gnats_cmd("LIST $listtype");
+  my $r = $self->_do_gnats_cmd("LIST $listtype");
 
-  if (not $self->_is_code_ok($code)) {
-    $self->_mark_error($code, $response);
+  if (not $self->_is_code_ok($r->code)) {
+    $self->_mark_error($r);
     return;
   }
 
-  my @result = qw();
-
-  my @rawrows = $self->_extract_list_content($response);
-  foreach my $row (@rawrows) {
-    push @result, map { @{ $keynames }[$_] =>
-                          $rawrows[$_] } 0..( scalar @{$keynames} - 1);
+  my $result = [];
+  foreach my $row (@{ $r->raw }) {
+    my @parts = split ':', $row;
+    push @{ $result}, { map { @{ $keynames }[$_] =>
+                                $parts[$_] } 0..( scalar @{$keynames} - 1) };
   }
-  return \@result;
+  return $result;
 }
 
 sub login {
@@ -1003,9 +1012,9 @@ sub login {
     $pass = q{*};
   }
 
-  my ($code, $response) = $self->_do_gnats_cmd("CHDB $db $user $pass");
+  my $r = $self->_do_gnats_cmd("CHDB $db $user $pass");
 
-  if ($self->_is_code_ok($code)) {
+  if ( $r->code == $CODE_OK ) {
     $self->{db}   = $db;
     $self->{user} = $user;
     $self->{pass} = $pass;
@@ -1013,10 +1022,44 @@ sub login {
     return 1;
   }
 
-  $self->_mark_error($code, $response);
-  return 0;
+  if ( $r->code == $CODE_NO_ACCESS ) {
+    logerror( 'ERROR: CODE_NO ACCESS: ' . $r->raw );
+    return;
+  }
+
+  if ( $r->code == $CODE_INVALID_DATABASE ) {
+    logerror( 'ERROR: CODE_NO ACCESS: ' . $r->raw );
+    return;
+  }
+
+  logerror( 'ERROR: LOGIN: UNKNOWN RESPONSE: ' . $r->raw );
+  return;
 }
 
+# Specify the user for database access.
+# A 350 is not returned in this case.
+sub cmd_user {
+  my ( $self, $user, $pass) = @_;
+
+  return if not defined $user or not defined $pass;
+
+  my $r = $self->_do_gnats_cmd("USER $user $pass");
+
+  if ( $r->code == $CODE_OK ) {
+    $self->_set_access_mode;
+    return 1;
+  }
+
+  if ( $r->code == $CODE_NO_ACCESS ) {
+    logerror( 'ERROR: CODE_NO_ACCESS: ' . $r->raw );
+    return
+  }
+
+  logerror( 'ERROR: LOGIN: UNKNOWN RESPONSE: ' . $r->raw );
+  return;
+}
+
+
 sub get_access_mode {
     my ( $self ) = @_;
     return $self->{accessMode};
@@ -1025,18 +1068,20 @@ sub get_access_mode {
 # This is called by login to determine the current access mode,
 # typically this would not be called by the user.
 sub _set_access_mode {
-    my ( $self )  = @_;
+  my ( $self )  = @_;
 
-    $self->{accessMode} = undef; # Clear it.
-    my ($code, $response) = $self->_do_gnats_cmd('USER');
-    if ($self->_is_code_ok($code)) {
-        $response =~ s/.*\n350\s*(\S+)\s*\n/$1/sxm;
-        $self->{accessMode} = $response;
-        return $response;
-    } else {
-        $self->_mark_error($code, $response);
-        return 0;
-    }
+  $self->{accessMode} = undef;
+
+  my $r = $self->_do_gnats_cmd('USER');
+
+  if ($self->_is_code_ok($r->code)) {
+    $self->{accessMode} = shift @{ $r->raw };
+    $self->{accessMode} =~ s/.*\n350\s*(\S+)\s*\n/$1/gsm;
+    return $self->{accessMode};
+  }
+
+  $self->_mark_error($r);
+  return 0;
 }
 
 
@@ -1061,7 +1106,7 @@ sub _do_gnats_cmd {
 
   my $r = $self->_process;
 
-  return ($r->code, $r->raw);
+  return $r;
 }
 
 sub _process {
@@ -1069,13 +1114,6 @@ sub _process {
 
   my $r = $self->_get_gnatsd_response;
 
-#  if ($r->code == $CODE_INVALID_ENUM) {
-#    while ( $self->_gsock->getline ) { debug( $_ ) ; }
-#  }
-
-  $self->{lastCode} = $r->code;
-  $self->{lastResponse} = $r->raw;
-
   return $r;
 }
 
@@ -1110,7 +1148,7 @@ sub _read {
 
   my $line = $self->_gsock->getline;
 
-  if ( not defined $line ) { return; }
+  if ( not defined $line ) { return $response; }
 
   $line = $self->_read_clean($line);
 
@@ -1120,7 +1158,7 @@ sub _read {
 
   $response->code( @{ $result }[0] );
 
-  if ( not defined $response->code ) { return; }
+  if ( $response->code == -1 ) { return $response; }
 
   if ( not ( $response->code == $CODE_PR_READY           or
              $response->code == $CODE_TEXT_READY         or
@@ -1165,8 +1203,8 @@ sub _read_clean {
   my ( $self, $line ) = @_;
   if ( not defined $line ) { return; }
 
-  $line =~ s/[\r\n]//gsxm;
-  $line =~ s/^[.][.]/./sxm;
+  $line =~ s/[\r\n]//gsm;
+  $line =~ s/^[.][.]/./gsm;
   return $line;
 }
 
@@ -1199,28 +1237,28 @@ sub _clear_error {
 
 
 sub _mark_error {
-  my ($self, $code, $msg) = @_;
+  my ($self, $r) = @_;
 
-  $self->{errorCode} = $code;
-  $self->{errorMessage} = $msg;
-  debug('ERROR: CODE: [' . $code . '] MSG: [' . $msg . ']');
+  $self->{errorCode} = $r->code;
+  $self->{errorMessage} = $r->raw;
+  debug('ERROR: CODE: [' . $r->code . '] MSG: [' . $r->raw . ']');
 
   return;
 }
 
 sub debug {
   my ( $message ) = @_;
-  if ( not defined $debug_gnatsd) { return; }
+  if ( not defined $debug_gnatsd or
+       not $debug_gnatsd) { return; }
   if ( not ( print 'DEBUG: [' . $message . ']' . $NL ) ) {
-    carp ( 'weird - could not print trace string' );
+    logerror ( 'weird - could not print trace string' );
   }
   return;
 }
 
-# preloaded methods go here.
-
-# Autoload methods go after =cut, and are processed by the autosplit program.
-
+sub logerror {
+  print shift . "\n";
+}
 
 
 1;
@@ -1269,11 +1307,11 @@ Net::Gnats - Perl interface to GNU Gnats daemon
 
 =head1 DESCRIPTION
 
-Net::Gnats provides a perl interface to the gnatsd command set.  Although 
+Net::Gnats provides a perl interface to the gnatsd command set.  Although
 most of the gnatsd command are present and can be explicitly called through
-Net::Gnats, common gnats tasks can be accompished through some methods 
+Net::Gnats, common gnats tasks can be accompished through some methods
 which simplify the process (especially querying the database, editing bugs,
-etc). 
+etc).
 
 The current version of Net::Gnats (as well as related information) is
 available at http://gnatsperl.sourceforge.net/
@@ -1414,6 +1452,15 @@ Constructor, optionally taking one or two arguments of hostname and
 port of the target gnats server.  If not supplied, the hostname
 defaults to localhost and the port to 1529.
 
+=head2 skip_version_check
+
+If you are using a custom gnats daemon, your version number might
+"not be supported".  If you are sure you know what you are doing
+and am willing to take the consequences:
+
+ my $g = Net::Gnats->new();
+ $g->skip_version_check(1);
+
 =head2 gnatsd_connect
 
 Connects to the gnats server.  No arguments.  Returns true if
